#! /usr/bin/env python
import sys
import os

from django.conf import settings, global_settings


def main():
    current_dir = os.path.dirname(__file__)
    app_name = os.path.basename(current_dir)
    sys.path.insert(0, os.path.join(current_dir, '..'))

    if not settings.configured:
        settings.configure(
            INSTALLED_APPS=('django.contrib.auth', 'django.contrib.contenttypes', 'django.contrib.admin', app_name),
            DATABASES={'default': {'ENGINE': 'django.db.backends.sqlite3'}},
            MIDDLEWARE_CLASSES=global_settings.MIDDLEWARE_CLASSES,  # Prevents Django 1.7 warning.
        )

    try:  # Django 1.7 +
        from django import setup
        setup()
    except ImportError:
        pass

    from django.test.utils import get_runner
    runner = get_runner(settings)()
    failures = runner.run_tests((app_name,))

    sys.exit(failures)


if __name__ == '__main__':
    main()



from django.conf import settings


# Toggles internal preferences model showing up in the Admin.
EXPOSE_MODEL_TO_ADMIN = getattr(settings, 'SITEPREFS_EXPOSE_MODEL_TO_ADMIN', False)

# Module name used by siteprefs.toolbox.autodiscover_siteprefs() to find preferences in application packages.
PREFS_MODULE_NAME = getattr(settings, 'SITEPREFS_MODULE_NAME', 'settings')

# List of commands called by manage.py where autodiscover is enabled
MANAGE_SAFE_COMMANDS = getattr(settings, "SITEPREFS_MANAGE_SAFE_COMMANDS",
                           ['runserver', 'runserver_plus', 'run_gunicorn', 'celeryd'])



from django.contrib import admin

from .settings import EXPOSE_MODEL_TO_ADMIN


if EXPOSE_MODEL_TO_ADMIN:
    from .models import Preference

    class PreferenceAdmin(admin.ModelAdmin):

        list_display = ('app', 'name')
        search_fields = ['app', 'name']
        list_filter = ['app']
        ordering = ['app', 'name']


    admin.site.register(Preference, PreferenceAdmin)



import os
import inspect
from datetime import datetime
from collections import OrderedDict

from django.db import models
from django.utils.translation import ugettext_lazy as _

try:
    from django.utils.module_loading import import_module as import_module_
except ImportError:
    # Django <=1.9.0
    from django.utils.importlib import import_module as import_module_

from django.utils.module_loading import module_has_submodule
from django.contrib import admin

from .signals import prefs_save
from .settings import PREFS_MODULE_NAME


class Frame(object):
    """Represents a frame object at a definite level of hierarchy.

    To be used as context manager:

        with Frame as f:
            ...

    """

    def __init__(self, stepback=0):
        self.depth = stepback

    def __enter__(self):
        frame = inspect.currentframe().f_back
        for __ in range(self.depth):
            frame = frame.f_back
        self.frame = frame
        return self.frame

    def __exit__(self, exc_type, exc_val, exc_tb):
        del self.frame


class PatchedLocal(object):
    """Object of this class temporarily replace all module variables
    considered preferences."""

    def __init__(self, key, val):
        self.key = key
        self.val = val


class PrefProxy(object):
    """Objects of this class replace app preferences."""

    def __init__(self, name, default, category=None, field=None, verbose_name=None, help_text='', static=True,
                 readonly=False):
        self.name = name
        self.category = category
        self.default = default
        self.static = static
        self.help_text = help_text
        if static:
            readonly = True
        self.readonly = readonly
        if verbose_name is None:
            verbose_name = name.replace('_', ' ').capitalize()
        self.verbose_name = verbose_name
        if field is None:
            self.field = get_field_for_proxy(self)
        else:
            self.field = field
            update_field_from_proxy(self.field, self)

    def get_value(self):
        if self.static:
            val = self.default
        else:
            try:
                val = getattr(self, 'db_value')
            except AttributeError:
                val = self.default
        return self.field.to_python(val)

    def __cmp__(self, other):
        if self.get_value() < other:
            return -1
        elif self.get_value() > other:
            return 1
        return 0

    def __call__(self, *args, **kwargs):
        return self.get_value()

    def __str__(self):
        return '%s' % self.get_value()

    def __repr__(self):
        return '%s = %s' % (self.name, self.get_value())


def get_field_for_proxy(pref_proxy):
    """Returns a field object instance for a given PrefProxy object."""
    field = {
        bool: models.BooleanField,
        int:  models.IntegerField,
        float: models.FloatField,
        datetime: models.DateTimeField,
    }.get(type(pref_proxy.default), models.TextField)()
    update_field_from_proxy(field, pref_proxy)
    return field


def update_field_from_proxy(field_obj, pref_proxy):
    """Updates field object with data from a PrefProxy object."""
    attr_names = ['verbose_name', 'help_text', 'default']
    for attr_name in attr_names:
        setattr(field_obj, attr_name, getattr(pref_proxy, attr_name))


def get_pref_model_class(app, prefs, get_prefs_func):
    """Returns preferences model class dynamically crated for a given app or None on conflict."""

    model_dict = {
            '_prefs_app': app,
            '_get_prefs': staticmethod(get_prefs_func),
            '__module__': '%s.%s' % (app, PREFS_MODULE_NAME),
            'Meta': type('Meta', (models.options.Options,), {
                'verbose_name': _('Preference'),
                'verbose_name_plural': _('Preferences')
            })
    }

    for field_name, val_proxy in prefs.items():
        model_dict[field_name] = val_proxy.field

    try:  # Make Django 1.7 happy.
        model = type('Preferences', (models.Model,), model_dict)
    except RuntimeError:
        return None

    def fake_save_base(self, *args, **kwargs):
        updated_prefs = {
            f.name: getattr(self, f.name) for f in self._meta.fields if not isinstance(f, models.fields.AutoField)
        }
        app_prefs = self._get_prefs(self._prefs_app)
        for pref in app_prefs.keys():
            if pref in updated_prefs:
                app_prefs[pref].db_value = updated_prefs[pref]
        self.pk = self._prefs_app  # Make Django 1.7 happy.
        prefs_save.send(sender=self, app=self._prefs_app, updated_prefs=updated_prefs)
        return True

    model.save_base = fake_save_base

    return model


def get_pref_model_admin_class(prefs):
    by_category = OrderedDict()
    readonly_fields = []

    for field_name, val_proxy in prefs.items():

        if val_proxy.readonly:
            readonly_fields.append(field_name)

        if val_proxy.category not in by_category:
            by_category[val_proxy.category] = []

        by_category[val_proxy.category].append(field_name)

    cl_model_admin_dict = {
        'has_add_permission': lambda *args: False,
        'has_delete_permission': lambda *args: False
    }

    if readonly_fields:
        cl_model_admin_dict['readonly_fields'] = readonly_fields

    fieldsets = []
    for category, cat_prefs in by_category.items():
        fieldsets.append((category, {'fields': cat_prefs}))
    if fieldsets:
        cl_model_admin_dict['fieldsets'] = fieldsets

    model = type('PreferencesAdmin', (admin.ModelAdmin,), cl_model_admin_dict)
    model.changelist_view = lambda self, request, **kwargs: self.change_view(request, '', **kwargs)
    model.get_object = lambda self, *args: (
        self.model(
            **{
                field_name: val_proxy.get_value() for field_name, val_proxy in
                self.model._get_prefs(self.model._prefs_app).items()
            }
        )
    )

    return model


def get_frame_locals(stepback=0):
    """Returns locals dictionary from a given frame."""
    with Frame(stepback=stepback) as frame:
        locals_dict = frame.f_locals
    return locals_dict


def traverse_local_prefs(stepback=0):
    """Generator to walk through variables considered as preferences
    in locals dict of a given frame.

    """
    locals_dict = get_frame_locals(stepback+1)
    for k in locals_dict:
        if not k.startswith('_') and k.upper() == k:
            yield k, locals_dict


def import_module(package, module_name):
    """Imports a module from a given package."""
    module = import_module_(package)
    try:
        import_module_('%s.%s' % (package, module_name))
    except:
        if module_has_submodule(module, module_name):
            raise


def import_prefs():
    """Imports preferences modules from packages (apps)."""
    settings_locals = get_frame_locals(3)

    project_package = settings_locals['__package__']  # Expected project layout introduced in Django 1.4
    if not project_package:
        # Fallback to old layout.
        project_package = os.path.split(os.path.dirname(settings_locals['__file__']))[-1]

    # Try to import project-wide prefs.
    import_module(project_package, PREFS_MODULE_NAME)

    from django.conf import settings

    for app in settings.INSTALLED_APPS:
        import_module(app, PREFS_MODULE_NAME)



from django.db import models
from django.db.utils import IntegrityError
from django.utils.translation import ugettext_lazy as _
from django.utils.encoding import python_2_unicode_compatible


@python_2_unicode_compatible
class Preference(models.Model):

    app = models.CharField(_('Application'), max_length=100, null=True, blank=True, db_index=True)
    name = models.CharField(_('Name'), max_length=150)
    text = models.TextField(_('Value'), null=True, blank=True)

    class Meta(object):
        verbose_name = _('Preference')
        verbose_name_plural = _('Preferences')
        unique_together = ('app', 'name')

    def __str__(self):
        return '%s:%s' % (self.app, self.name)

    @classmethod
    def read_prefs(cls, mem_prefs):
        """Initializes preferences entries in DB according to currently discovered prefs."""
        db_prefs = {
            '%s__%s' % (pref['app'], pref['name']): pref for pref in
            cls.objects.values().order_by('app', 'name')
        }

        new_prefs = []
        for app, prefs in mem_prefs.items():
            for pref_name, pref_proxy in prefs.items():
                if not pref_proxy.static:  # Do not add static options to DB.
                    key = '%s__%s' % (app, pref_name)
                    if key in db_prefs:
                        # Entry already exists in DB. Let's get pref value from there.
                        pref_proxy.db_value = db_prefs[key]['text']
                    else:
                        new_prefs.append(cls(app=app, name=pref_name, text=pref_proxy.default))

        if new_prefs:
            try:
                cls.objects.bulk_create(new_prefs)
            except IntegrityError:  # Don't bother with duplicates.
                pass

    @classmethod
    def update_prefs(cls, *args, **kwargs):
        # TODO That could be more efficient.
        updated_prefs = kwargs['updated_prefs']
        for db_pref in cls.objects.filter(app=kwargs['app']):
            if db_pref.name in updated_prefs:
                db_pref.text = updated_prefs[db_pref.name]
                db_pref.save()



from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class SiteprefsConfig(AppConfig):
    """The default siteprefs configuration."""

    name = 'siteprefs'
    verbose_name = _('Site Preferences')



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Preference',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('app', models.CharField(blank=True, max_length=100, verbose_name='Application', db_index=True, null=True)),
                ('name', models.CharField(max_length=150, verbose_name='Name')),
                ('text', models.TextField(blank=True, verbose_name='Value', null=True)),
            ],
            options={
                'verbose_name_plural': 'Preferences',
                'verbose_name': 'Preference',
            },
            bases=(models.Model,),
        ),
        migrations.AlterUniqueTogether(
            name='preference',
            unique_together=set([('app', 'name')]),
        ),
    ]



"""This file contains tests for siteprefs."""
import sys
import unittest

from django.contrib import admin
from django.db import models

from .utils import Frame, PatchedLocal, PrefProxy, get_field_for_proxy, get_pref_model_class, \
    get_pref_model_admin_class, get_frame_locals, import_module, PREFS_MODULE_NAME
from .toolbox import autodiscover_siteprefs


class FakeSettingsModule(object):

    __name__ = 'siteprefs.settings'


class ToolboxTest(unittest.TestCase):

    def test_autodiscover(self):
        __package__ = 'siteprefs'
        autodiscover_siteprefs()


class UtilsTest(unittest.TestCase):

    def test_frame(self):
        with Frame() as f:
            self.assertEqual(f.f_locals['self'], self)

    def test_patched_local(self):
        pl = PatchedLocal('k', 'v')
        self.assertEqual(pl.key, 'k')
        self.assertEqual(pl.val, 'v')

    def test_pref_poxy(self):
        pp = PrefProxy('proxy_name', 'default')
        self.assertEqual(pp.name, 'proxy_name')
        self.assertEqual(pp.default, 'default')
        self.assertIsNone(pp.category)
        self.assertIsInstance(pp.field, models.TextField)
        self.assertEqual(pp.verbose_name, 'Proxy name')
        self.assertEqual(pp.help_text, '')
        self.assertTrue(pp.static)
        self.assertTrue(pp.readonly)
        self.assertEqual(pp.get_value(), 'default')

        pp = PrefProxy('proxy_name_2', 2, category='cat', field=models.IntegerField(), verbose_name='verbose name', help_text='help text', static=False, readonly=True)
        self.assertEqual(pp.name, 'proxy_name_2')
        self.assertEqual(pp.default, 2)
        self.assertEqual(pp.category, 'cat')
        self.assertIsInstance(pp.field, models.IntegerField)
        self.assertEqual(pp.verbose_name, 'verbose name')
        self.assertEqual(pp.help_text, 'help text')
        self.assertFalse(pp.static)
        self.assertTrue(pp.readonly)
        self.assertEqual(pp.get_value(), 2)

        pp.db_value = 42
        self.assertEqual(pp.get_value(), 42)
        self.assertEqual(pp(), 42)
        self.assertTrue(pp() < 43)
        self.assertTrue(pp() > 41)

        pp = PrefProxy('proxy_name_3', 10, static=False)
        self.assertFalse(pp.readonly)

    def test_get_field_for_proxy(self):
        pp = PrefProxy('proxy_name', 10)
        self.assertIsInstance(get_field_for_proxy(pp), models.IntegerField)

        pp = PrefProxy('proxy_name', True)
        self.assertIsInstance(get_field_for_proxy(pp), models.BooleanField)

        pp = PrefProxy('proxy_name', 13.4)
        self.assertIsInstance(get_field_for_proxy(pp), models.FloatField)

        pp = PrefProxy('proxy_name', 'abc')
        self.assertIsInstance(get_field_for_proxy(pp), models.TextField)

    def test_get_pref_model_class(self):
        p1 = PrefProxy('pp1', 10)
        p2 = PrefProxy('pp2', 20)
        p3 = PrefProxy('pp3', 'admin', field=models.CharField(max_length=10))
        p4 = PrefProxy('pp4', 'another', verbose_name='verbosed', static=False, field=models.CharField(max_length=10))

        my_prefs_func = lambda: 'yes'

        sys.modules['siteprefs.settings'] = FakeSettingsModule()

        cl = get_pref_model_class('siteprefs', {'pp1': p1, 'pp2': p2, 'pp3': p3, 'pp4': p4}, my_prefs_func)
        model = cl()
        self.assertTrue(issubclass(cl, models.Model))
        self.assertIsInstance(model._meta.fields[3], models.CharField)
        self.assertEquals(model._meta.fields[4].verbose_name, 'verbosed')

    def test_get_pref_model_admin_class(self):

        p1 = PrefProxy('pp1', 10)
        p2 = PrefProxy('pp2', 20)

        cl = get_pref_model_admin_class({'pp1': p1, 'pp2': p2})
        self.assertTrue(issubclass(cl, admin.ModelAdmin))

    def test_get_frame_locals(self):
        a = 'a'
        b = 'b'
        l = get_frame_locals(1)
        self.assertIn('a', l)
        self.assertIn('b', l)

    def test_import_module(self):
        try:
            import_module('siteprefs', PREFS_MODULE_NAME)
        except Exception:
            self.fail('test_import_module failed')



# -*- coding: utf-8 -*-
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding model 'Preference'
        db.create_table('siteprefs_preference', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('app', self.gf('django.db.models.fields.CharField')(null=True, db_index=True, blank=True, max_length=100)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=150)),
            ('text', self.gf('django.db.models.fields.TextField')(null=True, blank=True)),
        ))
        db.send_create_signal('siteprefs', ['Preference'])

        # Adding unique constraint on 'Preference', fields ['app', 'name']
        db.create_unique('siteprefs_preference', ['app', 'name'])


    def backwards(self, orm):
        # Removing unique constraint on 'Preference', fields ['app', 'name']
        db.delete_unique('siteprefs_preference', ['app', 'name'])

        # Deleting model 'Preference'
        db.delete_table('siteprefs_preference')


    models = {
        'siteprefs.preference': {
            'Meta': {'object_name': 'Preference', 'unique_together': "(('app', 'name'),)"},
            'app': ('django.db.models.fields.CharField', [], {'null': 'True', 'db_index': 'True', 'blank': 'True', 'max_length': '100'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '150'}),
            'text': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'})
        }
    }

    complete_apps = ['siteprefs']